from .edges import Edge, START, END
from .nodes import GraphNode
from .states import State, Stream, StateUpdate
from .graph import GraphExecutor

__all__ = [
    "Edge",
    "GraphNode",
    "State",
    "Stream",
    "StateUpdate",
    "GraphExecutor",
    "START",
    "END",
]


from typing import Callable, Type, TypeVar, Generic
from .states import State
from .nodes import GraphNode


class START:
    pass

class END:
    pass


T = TypeVar('T', bound=State)

class Edge(Generic[T]):

    source: GraphNode[T] | Type[START] | list[GraphNode[T] | Type[START]]
    next: Callable[[T], GraphNode[T] | Type[END]]

    def __init__(self, source: GraphNode[T] | Type[START] | list[GraphNode[T] | Type[START]], next: Callable[[T], GraphNode[T] | Type[END]]):
        self.source = source
        self.next = next



from .edges import Edge, START
from .nodes import GraphNode
from .states import State, StateUpdate
from typing import Type, TypeVar, Generic, Callable, Awaitable, Sequence
from collections import defaultdict
import asyncio

T = TypeVar('T', bound=State)

class GraphExecutor(Generic[T]):

    edges: list[Edge[T]]

    def __init__(self, edges: list[Edge[T]]):
        self.edges = edges
        self._index: dict[GraphNode[T] | Type[START], list[Edge[T]]] = defaultdict(list[Edge[T]])

        for edge in self.edges:
            if isinstance(edge.source, list):
                for source in edge.source:
                    self._index[source].append(edge)
            else:
                self._index[edge.source].append(edge)


    async def __call__(self, initial_state: T) -> T:
        state: T = initial_state
        current_nodes: Sequence[GraphNode[T] | Type[START]] = [START]

        while True:

            edges: list[Edge[T]] = []

            for current_node in current_nodes:

                # Find the edge corresponding to the current node
                edges.extend(self._index[current_node])


            next_nodes: list[GraphNode[T]] = [
                n for edge in edges 
                if isinstance((n := edge.next(state)), GraphNode)
            ] # Only one execution of next filtering "END"s


            if not next_nodes:
                break # END


            parallel_tasks: list[Callable[[T], Awaitable[list[StateUpdate[T]]]]] = []
            # Determine the next node using the edge's next function
            for next_node in next_nodes:
                
                parallel_tasks.append(next_node.run)

            # Run parallel
            update_tasks_list: list[list[StateUpdate[T]]] = await asyncio.gather(*(task(state) for task in parallel_tasks))

            for update_tasks in update_tasks_list:
                for update in update_tasks:
                    update(state)

            
            current_nodes = next_nodes


        return state


from abc import ABC, abstractmethod
from pydantic import BaseModel, Field, ConfigDict
from typing import TypeVar, AsyncIterator, Generic, Protocol, Any
from types import TracebackType


T = TypeVar('T', covariant=True, default=object)

class Stream(ABC, AsyncIterator[T]):

    @abstractmethod
    async def aclose(self) -> None:
        pass

    @abstractmethod
    async def __anext__(self) -> T:
        pass

    async def __aenter__(self) -> "Stream[T]":
        return self

    async def __aexit__(
            self, exc_type: type[BaseException] | None, 
            exc: BaseException | None, 
            tb: TracebackType | None
        ) -> None: # Not handling exceptions here -> returns None

        await self.aclose()


class State(BaseModel, Generic[T]):
    vars: dict[str, object] = Field(default_factory=dict)
    streams: dict[str, 'Stream[T]'] = Field(default_factory=dict)

    model_config = ConfigDict(arbitrary_types_allowed=True)


S = TypeVar('S', bound=State, contravariant=True)

class StateUpdate(Protocol, Generic[S]):
    def __call__(self, s: S) -> Any: ...